/*
Name:   graph.h
Zweck:  Aufnahme der Vorabdefinitionen von graph.c
Autor:  Andreas Seifert
Datum:
Update:
*/


#ifndef _GRAPH_H_
#define _GRAPH_H_


/********************************************************/
/*  graph.h - header file fuer graph.c                  */
/********************************************************/


/************* Allgemeine Header-Dateien einbinden **************************/

#include <stdio.h>  /* Standard-Header-Datei fuer C */
#include <Xm/Xm.h>  /* OSF/Motif-Widget-Set */


#include "struct.h" /* enthaelt die Definitionen der Datenstrukturen */
#include "globals.h" /* enthaelt globale Variablen */



/* Vorabdefinitionen */
NODE         *create_Graph();
NODE         *correspondingNetNode(char [42]);
GRAPHIC_NODE *correspondingGraphicNode(char [42]);
NODE         *addNetNode(char [4], char[40], double);
void         insertNetEdge(NODE *, NODE *, double);
void         clear_Graph();
void         clear_GraphicObjects();
void         remove_arcs(GRAPHIC_NODE *);
void         remove_ident(GRAPHIC_NODE *);
Bool         deleteArc(GRAPHIC_NODE *, GRAPHIC_NODE *);
Bool         insertArc(GRAPHIC_NODE *, GRAPHIC_NODE *);
GRAPHIC_ARC  *arc_already_exists(GRAPHIC_NODE *, GRAPHIC_NODE *);
int          setGraphicArcWeight(GRAPHIC_NODE *, GRAPHIC_NODE *, int, int);
void         change_GraphicObject(int, int, GRAPHIC_NODE *);
GRAPHIC_NODE *node_hit(int, int);
void         add_GraphicObject(char [40], char[40], int);
void         addRepresentatives(GRAPHIC_NODE *, GRAPHIC_IDENT *);
void         add_GraphicVoteQuor(GRAPHIC_NODE *, int, int, int, int, int, int);
void         compute_Connections(int, int, GRAPHIC_NODE *);
Bool         validPosition(int, int, GRAPHIC_NODE *);
int          mark_version(GRAPHIC_IDENT *);
int          listcmp(GRAPHIC_IDENT *, GRAPHIC_IDENT *);  

/* Funktionen fuer graph.c */
GRAPHIC_NODE *findRoot(GRAPHIC_NODE *);
GRAPHIC_NODE *searchRoot(GRAPHIC_NODE *);
GRAPHIC_NODE *checkVQM(GRAPHIC_NODE *);
void         sortArc(GRAPHIC_ARC *, int);  
void         exchangeArcFromTo(GRAPHIC_ARC *, GRAPHIC_ARC *);
GRAPHIC_IDENT *createIdent(char [5]);

/* Funktionen fuer set.c */
GRAPHIC_SET           *createHeadOfNodeSet(GRAPHIC_NODE *, int);
REF_SET_ELEMENT       *getRefListElement(GRAPHIC_NODE *);
void                  addNodeSetElement(GRAPHIC_SET *, GRAPHIC_NODE *);
REF_SET_ELEMENT       *addObjectToRefList(REF_SET_ELEMENT *, GRAPHIC_NODE *);
void                  addObjectSetToRefList(REF_SET_ELEMENT *, GRAPHIC_SET *);
void                  removeObjectSetFromRefList(GRAPHIC_SET *);
void                  removeObjectFromRefList(GRAPHIC_NODE *);
void                  clean_VQM();
void                  deleteVQMSet(VQM_SET *);
void                  clean_ListOfReferences();
void                  clean_SuccessfulSons();
void                  addSolutionToVQMSet(int);
VQM_SET               *createVQMSet();   
VQM_SET_ELEMENT       *createVQMSetElement();
void                  insertElementIntoVQMSet(VQM_SET *, GRAPHIC_NODE *);
void                  insertItemIntoVQMSet(VQM_SET *, char [5]);
Bool                  existVQMSet(VQM_SET *, int);
Bool                  listcmpVQMSet(VQM_SET_ELEMENT *, VQM_SET_ELEMENT *);
Bool                  notMemberOfSet(VQM_SET *, char [5]);


#endif /* _GRAPH_H_ */


