/*
Name:   main.c
Zweck:  Verarbeiten der Benutzeraktionen, Kontrolle des Dialogs und
        Anzeige der Verarbeitungsergebnisse
Autor:  Andreas Seifert
Datum:
Update:
*/

/*--------------------------------------------------------------------------*/ 


/****************************************************************************/

#include <ctype.h>
#include "main.h"
#include "graph.h"
#include "draw.h"
#include "motif.h"
#include "file.h"
#include "set.h"
#include "compute.h"

/****************************************************************************/

/*
 * AU: Verarbeiten der Events in Callbacks und Koordinierung der
 *     Verarbeitung
 * ST: MenuCB()                     Menuepunkte des Dateimenues
 *     fileSelectCB()               Dialog zur Auswahl der Netztopologie- 
 *                                  und Voting-Struktur-Dateien
 *     nodeNameDialogCB()           Dialog zur Eingabe des Namens eines 
 *                                  virtuellen Rechners
 *     ZoomCB()                     Veraenderung der Zoomoption
 *     activateZoom()               Aktivierung der Zoomoption
 *     EditCB()                     Aktionen zur Manipulation der 
 *                                  Voting-Struktur
 *     installEventHandler()        Arbeitsbereich sensitivieren 
 *     set_sensitive()              Sensitivitaet der Knoepfe steuern
 *     mapScales()                  Anzeige Schieberegler
 *     clear_Nodelist()             Auswahlliste loeschen
 *     fill_Nodelist()              Auswahlliste fuellen
 *     listCB()                     Selektion in der Auswahlliste
 *     setNodeListSensitive()       Auswahlliste de- und sensitivieren
 *     start_rubber_band()          Rechnerumriss am Ausgangspunkt zeichnen
 *     interrupt_rubber_band()      Rechnerumriss beim Verlassen des
 *                                  Arbeitsbereichs zeichnen
 *     track_rubber_band()          Rechnerumriss beim Verschieben zeichnen
 *     end_nodeinsert_rubber_band() Positionierung eines einzufuegenden
 *                                  Rechners
 *     markListobject()             markiere Rechner in der Auswahlliste
 *     markListPos()                suche Rechner in der Auswahlliste
 *     point_in_delnode()           Rechner der Voting-Struktur loeschen
 *     restore_Listobject()         Rechner der Auswahlliste wieder 
 *                                  selektierbar 
 *     point_in_movenode()          Rechner der Voting-Struktur verschieben
 *     point_in_modifynode()        Wertebelegung Rechner veraendern
 *     leave_window_node()          Rechnerumriss beim Verlassen des
 *                                  Arbeitsbereichs zeichnen
 *     end_move_node()              Rechnerumriss am Ende des Verschiebens
 *                                  zeichnen
 *     deleteArcCB()                Verbindung zwischen Rechnern loeschen
 *     modify_arcCB()               Wertebelegung Verbindung veraendern
 *     insert_arcCB()               Verbindung zwischen Rechnern einfuegen
 *     modifyCB()                   Neue Wertebelegungen fuer Rechner und
 *                                  Verbindungen uebernehmen
 *     scaleCB()                    Schieberegler zur Veraenderung der
 *                                  Wertebelegungen ablesen
 *     FetchWorkingDirectory()      Bezeichnung Arbeitsverzeichnis ermitteln
 *     fileNameDialogCB()           Dateinamen eingeben
 *     store_freeGraphicCB()        Voting-Struktur abspeichern
 *     getDrawingSize()             Groesse Arbeitsbereich ermitteln
 *     checkDrawingSize()           Groesse  Arbeitsbereich anpassen
 *     exposeCB()                   Inhalt Arbeitsbereich herstellen
 *     validNodeName()              Bezeichnung eines virtuellen Rechners
 *                                  pruefen
 *     resultNodeNameCB()           Dialog zur Eingabe der Bezeichnung des
 *                                  virtuellen Rechners
 *     extractListString()          Rechnername aus einem Listeneintrag
 *                                  ermitteln
 *     ascending()                  Aufsteigende Reihenfolge ermitteln
 *     printUsage()                 Hinweis auf den Gebrauch der 
 *                                  Kommandozeilenoptionen
 *     RunCB()                      Menuepunkte des Analysemenues
 *     FetchWorkingDirectory()      Arbeitsverzeichnis und Rechner ermitteln
 *     showResultsCB()              Anzeige der Berechnungsergebnisse beenden
 *     ActualPercent()              Ausgabe des aktuellen Bearbeitungsstandes
 *     printStatus()                Ausgabe Groesse Arbeitsbereich
 */

 
/****************************************************************************/
/****************                             *******************************/
/****************    Beginn der Funktionen    *******************************/
/****************                             *******************************/
/****************************************************************************/




/*ARGSUSED*/
void MenuCB (widget, client_data, call_data)
   Widget   widget;      /* ausloesendes Widget */
   caddr_t  client_data, call_data;   
/*
 * AU: Betaetigung der Menuepunkte des Dateimenues verarbeiten
 * ST: clear_Graph();
 *     clear_GraphicObjects();
 *     clean_Workbench();
 */
{
   char     *ptrchr;
   XmString filter_txt; /* Dateifiltertext */
   static   Widget speichernAlsText = NULL;


   switch (*((int*)client_data))
   {
          case NEW_TREE_BUTTON : /* Neue Voting-Struktur */
               XmScaleSetValue(Widgetliste[FILE_NAME_SCALE], 5);
               XtPopup(Widgetliste[FILE_NAME_SHELL], XtGrabExclusive); 
               break;

          case OPEN_NET_BUTTON : /* Netztopologie oeffnen */
               filter_txt = 
                     XmStringCreateLtoR("*.top", charset);
               XtVaSetValues(Widgetliste[FILE_SELECT], 
                             XmNdirMask, filter_txt, 
                             NULL);        
               XtManageChild(Widgetliste[FILE_SELECT]);
               XmStringFree(filter_txt);
               break;

          case OPEN_TREE_BUTTON : /* Voting-Struktur oeffnen */
               filter_txt = 
                     XmStringCreateLtoR("*.vot", charset);
               XtVaSetValues(Widgetliste[FILE_SELECT],      
                             XmNdirMask, filter_txt, 
                             NULL);
               XtManageChild(Widgetliste[FILE_SELECT]);
               XmStringFree(filter_txt);
               break;
            
          case SAVE_YN_BUTTON : /* Voting-Struktur sichern */
               if (speichernAlsText == NULL)  
                  speichernAlsText = 
                      XmSelectionBoxGetChild(Widgetliste[FILE_NAME_PROMPT], 
                                             XmDIALOG_TEXT);    

               ptrchr = (char *)strrchr(data.filename_VOT, SLASH);

               if (ptrchr == NULL)
                  XmTextSetString(speichernAlsText, data.filename_VOT);
               else XmTextSetString(speichernAlsText, ptrchr+1);

               XtManageChild(Widgetliste[FILE_NAME_PROMPT]);
               break; 

          case HELP_MENU_BUTTON : /* Hilfemenue */
               XtManageChild(Widgetliste[HELP_MENU_MESSAGE]);
               break;

          case CLOSE_BUTTON : /* Programm beenden */
               clear_Graph();
               clean_SuccessfulSons();
               clear_GraphicObjects();
               clean_Workbench();
               clean_VQM();
               clean_ListOfReferences();
               exit(GOOD);
               break;

          default:
               XBell(display, 0);
               wprint(unknown_event, "main.c : MenuCB");
               break;
   }
}


/*ARGSUSED*/
void file_SelectCB(widget, clientDaten, SBinfos)
     Widget                       widget;      
     caddr_t                      clientDaten; /* hier nicht verwendet */
     XmSelectionBoxCallbackStruct *SBinfos; 
/*
 * AU: Ergebnis der Dateiauswahl verarbeiten
 * ST: readFileSuffix()
 *     create_Graph()
 *     loadObjectsFile()
 *     clear_GraphicObjects()
 *     clearArea()
 *     redrawArea()
 *     clear_Nodelist()
 *     fill_Nodelist()
 *     checkDrawingSize()
 *     clear_Graph()
 */
{
  String currentFileName;
  char   *ptrchr;

  switch(SBinfos->reason)
  {
        case XmCR_OK:
             XmStringGetLtoR (SBinfos->value, charset, &currentFileName);  

             switch(readFileSuffix(currentFileName))
             {
                   case top_file: /* Netztopologiedatei */
                        if (ptrNetObjectListFront != NULL) clear_Graph();

                        wprint(load_net_file, currentFileName); 
                        strncpy(data.filename_TOP, currentFileName, 255);
                        ptrchr = (char *)strstr(data.filename_TOP, ".");

                        if (ptrchr != NULL) *ptrchr = EOS;

                        ptrNetObjectListFront = (NODE *)create_Graph();

                        if (!loadObjectsFile(data.filename_TOP, top_file)) {
                           XBell(display, 0);
                           wprint(loadingNetObjectsfailed);
                           clear_Graph();
                           }
                        else { /* Fehler beim Einlesen, daher wieder Initialisierung */
                             clear_GraphicObjects();
                             clearArea();
                             redrawArea(ptrGraphicObjectList);
                             clear_Nodelist();
                             fill_Nodelist();
                             XtSetSensitive(Widgetliste[NEW_TREE_BUTTON], 
                                            True);
                             XtSetSensitive(Widgetliste[OPEN_TREE_BUTTON], 
                                            True);
                             XtSetSensitive(Widgetliste[OPEN_NET_BUTTON], 
                                            False);
                             }
                        break;

                   case vot_file: /* Voting-Struktur-Datei */
                        if (ptrGraphicObjectList != NULL) {
                           clear_GraphicObjects();

                           if (zoomOpportunity) { /* Zoomoption */
                              zoomOption = ZOOM_100;
                              XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
                              XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
                              XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
			      }

                           clearArea(); 
                           redrawArea(ptrGraphicObjectList);
                           clear_Nodelist();
                           fill_Nodelist();
                           }

                        if (ptrNetObjectListFront == NULL) {
                           XBell(display, 0);
                           wprint(missingList);
                           break;
                           }

                        wprint(load_tree_file, currentFileName);
                        strncpy(data.filename_VOT, currentFileName, 255);
                        ptrchr = (char *)strstr(data.filename_VOT, ".");

                        if (ptrchr != NULL) *ptrchr = EOS;

                        if (!loadObjectsFile(data.filename_VOT, vot_file)) {
                           XBell(display, 0);
                           wprint(loadingGraphicObjectsfailed);
                           clean_SuccessfulSons();
                           clear_GraphicObjects();

                           if (zoomOpportunity) {
                              zoomOption = ZOOM_100;
                              XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
                              XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
                              XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
                              }

                           clearArea(); 
                           redrawArea(ptrGraphicObjectList);
                           }
                        else { /* Grafikdatendatei mit oeffnen */
                             strncpy(data.filename_DRW, currentFileName, 255);
                             ptrchr = (char *)strstr(data.filename_DRW, ".");

                             if (ptrchr != NULL) *ptrchr = EOS;     

                             if (!loadObjectsFile(data.filename_DRW,drw_file)) {
                                XBell(display, 0);
                                wprint(loadingDrawObjectsfailed);
                                clean_SuccessfulSons();
                                clear_GraphicObjects();

                                if (zoomOpportunity) {
                                   zoomOption = ZOOM_100;
                                   XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
                                   XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
                                   XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
                                   }

                                clearArea();
                                redrawArea(ptrGraphicObjectList);
                                }
                             else {
                                  checkDrawingsize();

                                   if (zoomOpportunity) {
                                      zoomOption = ZOOM_100;
                                      XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
                                      XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
                                      XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
				      }

                                  clearArea();
                                  redrawArea(ptrGraphicObjectList);            
                                  XtMapWidget(Widgetliste[EDIT_BUTTONS]);
                                  XtMapWidget(Widgetliste[SCROLLED_DRAW]);
                                  XtSetSensitive(Widgetliste[NEW_TREE_BUTTON], 
                                                 False);
                                  XtSetSensitive(Widgetliste[OPEN_TREE_BUTTON],
                                                 False);
                                  XtSetSensitive(Widgetliste[OPEN_NET_BUTTON],
                                                 False);
                                  XtSetSensitive(Widgetliste[SAVE_YN_BUTTON], 
                                                 True);
                                  XtSetSensitive(Widgetliste[CHECK_VQM_BUTT], 
                                                 True);
                                  }
                             } 
                        break;

                   case no_file: /* Unbekannter Dateityp */
                        XBell(display, 0);
                        wprint(strange_file);
                        break;

                   default:
                        XBell(display, 0);
                        wprint(unknown_event, "main.c: part 1 file_SelectCB");
                        break;
             }
             XtFree(currentFileName);
             break;

        case XmCR_HELP: /* Hilfetaste */
             XtManageChild(Widgetliste[HELP_FILE_MESSAGE]);
             break;

        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c: part 2 file_SelectCB");
             break;
  }
}     


/*ARGSUSED*/
void nodeNameDialogCB (widget, clientDaten, SBinfos)
     Widget                       widget;      
     caddr_t                      clientDaten; 
     XmSelectionBoxCallbackStruct *SBinfos;
/*
 * AB: Bezeichnung eines virtuellen Rechners lesen und ggf.
 *     Arbeitsbereich sensitivieren
 * ST: validNodeName()
 *     installEventHandler()
 *     setNodeListSensitive() 
 */
{
  char     *nodename; 
  XmString spaceText = XmStringCreateLtoR("R", charset);

  switch (SBinfos->reason)
  {
         case XmCR_OK : /* Eingabe einer Bezeichnung */

              XmStringGetLtoR(SBinfos->value, charset, &nodename);
              strncpy(data.title, nodename, 40);

              if (validNodeName(data.title)) { /* gueltige Bezeichnung */
                 strncpy(data.identity, data.title, 40);
                 strncpy(data.name, virtual_node, 40); 
                 data.virtualNode = True;
                 installEventHandler(NODE_INS_BUTTON, True);
                 XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                 XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                 }
              else XtManageChild(Widgetliste[NODE_MESSAGE]); /* Fehler */
              break;

        case XmCR_HELP : /* Eingabefeld loeschen */

             XtVaSetValues(widget,
                          XmNtextString, spaceText, 
                          NULL);
             XmStringFree(spaceText);
             XtManageChild(Widgetliste[NODE_NAME_PROMPT]);
             break;

        case XmCR_CANCEL : /* Eingabe abbrechen */

             setNodeListSensitive(True);        
             break;

        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c : nodeNameDialogCB");            
             break;
  }
}



/*ARGSUSED*/
void ZoomCB (widget, clientDaten, aufrufDaten)
   Widget   widget;
   caddr_t  clientDaten, aufrufDaten;
/*
 * AU: Auswahl einer Zoomoption
 * ST: clearArea()
 *     redrawArea()
 */
{
   switch (*((int*)clientDaten))
   {
          case ZOOM_100_BUTTON : /* normale Darstellung */
               XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
               XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
               XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
               zoomOption = ZOOM_100;
               break;

          case ZOOM_75_BUTTON : /* 75 % Darstellung */
               XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], False);
               XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], True);
               XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
               zoomOption = ZOOM_75;
               break;

          case ZOOM_50_BUTTON : /* 50 % Darstellung */
               XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], False);
               XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], True);
               XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
               zoomOption = ZOOM_50;
               break;

          default :
               XBell(display, 0);
               wprint(unknown_event, "main.c : ZoomCB");
               break;
   }

   /* Neue Darstellung anzeigen */
   clearArea();
   redrawArea(ptrGraphicObjectList);
}


void activateZoom()
/*
 * AU: Aktivierung der Zoomoption
 * ST: loadZoomFonts()
 */
{
   if (loadZoomFonts()) { /* Zeichensaetze geladen ? */
      XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);  
      XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
      fprintf(stderr, zoomaktiv); 
      zoomOpportunity = True;
      }
   else {
        XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
        fprintf(stderr, zoomfailed);
        zoomOpportunity = False;
        }
}


/*ARGSUSED*/
void EditCB (widget, clientDaten, aufrufDaten)
   Widget   widget;      /* ausloesendes Widget */
   caddr_t  clientDaten, aufrufDaten; 
/*
 * AU: Aktivierung der Aktionen zur Manipulation der Voting-Struktur
 * ST: moveCursor()
 *     set_sensitive()
 *     installEventHandler()
 */
{
   static Bool firsttime = True;

   static int  activated_button;

   switch (*((int*)clientDaten))
   {
          case NODE_INS_BUTTON: /* Rechner einfuegen */
               if (firsttime) {
                   firsttime = False;
                   moveCursor(0, 350);
                   set_sensitive(NODE_INS_BUTTON, False);
                   XtMapWidget(Widgetliste[NODE_LIST_CONT]);
                   activated_button = NODE_INS_BUTTON;
                   }
               break;

          case NODE_MOVE_BUTTON: /* Rechner verschieben */
               if (firsttime) {
                   firsttime = False;
                   set_sensitive(NODE_MOVE_BUTTON, False);
                   activated_button = NODE_MOVE_BUTTON;
                   installEventHandler(NODE_MOVE_BUTTON, True);
                   }
               break;

          case NODE_DEL_BUTTON: /* Rechner entfernen */
               if (firsttime) {
                   firsttime = False;
                   set_sensitive(NODE_DEL_BUTTON, False);
                   activated_button = NODE_DEL_BUTTON;
                   installEventHandler(NODE_DEL_BUTTON, True);
                   }
               break;

          case NODE_INFO_BUTTON: /* Rechner veraendern */
               if (firsttime) {
                   firsttime = False; 
                   set_sensitive(NODE_INFO_BUTTON, False);
                   activated_button = NODE_INFO_BUTTON;
                   installEventHandler(NODE_INFO_BUTTON, True);
                   }
               break;

          case ARC_INS_BUTTON: /* Beziehung einfuegen */
               if (firsttime) { 
                   firsttime = False;
                   set_sensitive(ARC_INS_BUTTON, False);
                   activated_button = ARC_INS_BUTTON;
                   installEventHandler(ARC_INS_BUTTON, True);
                   }
               break;

          case ARC_DEL_BUTTON: /* Beziehung entfernen */
               if (firsttime) {
                   firsttime = False;
                   set_sensitive(ARC_DEL_BUTTON, False);
                   activated_button = ARC_DEL_BUTTON;
                   installEventHandler(ARC_DEL_BUTTON, True);
                   }
               break;

          case ARC_INFO_BUTTON: /* Beziehung veraendern */
               if (firsttime) {
                   firsttime = False;
                   set_sensitive(ARC_INFO_BUTTON, False);
                   activated_button = ARC_INFO_BUTTON;
                   installEventHandler(ARC_INFO_BUTTON, True);
                   }
               break;

          case END_FUNC_BUTTON: /* Sensitivitaet des Arbeitsbereiches
                                 * entfernen */
               if (!firsttime) {
                   firsttime = True;
                   set_sensitive(END_FUNC_BUTTON, True);
                   
                   switch (activated_button) /* abhaengig von der 1. Auswahl */
                   {
                          case NODE_INS_BUTTON:
                               XtUnmapWidget(Widgetliste[NODE_LIST_CONT]); 
                               if (data.callbacks_added) 
                                   installEventHandler(NODE_INS_BUTTON, False);
                               break;
    
                          case NODE_MOVE_BUTTON:
                               installEventHandler(NODE_MOVE_BUTTON, False);
                               break;

                          case NODE_DEL_BUTTON:
                               installEventHandler(NODE_DEL_BUTTON, False);
                               break;

                          case NODE_INFO_BUTTON:
                               installEventHandler(NODE_INFO_BUTTON, False);
                               break;
                        
                          case ARC_INS_BUTTON:
                               installEventHandler(ARC_INS_BUTTON, False);
                               break;
                        
                          case ARC_DEL_BUTTON:
                               installEventHandler(ARC_DEL_BUTTON, False);
                               break;

                          case ARC_INFO_BUTTON:
                               installEventHandler(ARC_INFO_BUTTON, False);
                               break;              
                     
                          default:
                               XBell(display, 0);
                               wprint(unknown_event, "main.c : part 1 EditCB");
                               break;  
                   }
                   } 
               break;

          default:
               XBell(display, 0);
               wprint(unknown_event, "main.c : part 2 EditCB");
               break;
   }
} 



void installEventHandler(touched_Function, addEvent)
     int  touched_Function; /* ausgewaehlte Funktion */
     Bool addEvent; /* Sensitivitaet entfernen oder hinzufuegen */
{
     switch (touched_Function) 
     {
            case NODE_INS_BUTTON : /* Rechner einfuegen */
                 if (addEvent) { /* neue Sensitivitaet */
                    XtAddEventHandler(Widgetliste[DRAW_AREA],  
                                      EnterWindowMask, False,
                                      (XtEventHandler)start_rubber_band, NULL);
                    XtAddEventHandler(Widgetliste[DRAW_AREA], 
                                      PointerMotionMask, False,
                                      (XtEventHandler)track_rubber_band, NULL);
                    XtAddEventHandler(Widgetliste[DRAW_AREA],  
                                      LeaveWindowMask, False,
                                      (XtEventHandler)interrupt_rubber_band, NULL);
                    XtAddEventHandler(Widgetliste[DRAW_AREA],
                                      ButtonPressMask, False,
                                      (XtEventHandler)end_nodeinsert_rubber_band, NULL);
                    data.callbacks_added = True;
                    XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                    XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                    }
                 else { /* Sensitivitaet entfernen */
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],  
                                           EnterWindowMask, False,
                                           (XtEventHandler)start_rubber_band, NULL);
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],  
                                           LeaveWindowMask, False,
                                           (XtEventHandler)interrupt_rubber_band, NULL);
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],
                                           PointerMotionMask, False,
                                           (XtEventHandler)track_rubber_band, NULL);
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],
                                           ButtonPressMask, False,
                                           (XtEventHandler)end_nodeinsert_rubber_band, NULL);
                      data.callbacks_added = False;
                      XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                      XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                      setNodeListSensitive(True);
                      }
                 break;

            case NODE_MOVE_BUTTON: /* Rechner verschieben */
                 if (addEvent) { /* neue Sensitivitaet */ 
                    XtAddEventHandler(Widgetliste[DRAW_AREA],
                                      ButtonPressMask, False,
                                      (XtEventHandler)point_in_movenode, NULL);
                    XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                    XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                    }
                 else { /* Sensitivitaet entfernen */
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],
                                           ButtonPressMask, False,
                                           (XtEventHandler)point_in_movenode, NULL);
                      XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                      XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                      }
                 break;
          
            case NODE_DEL_BUTTON: /* Rechner entfernen */
                 if (addEvent) { /* neue Sensitivitaet */
                    XtAddEventHandler(Widgetliste[DRAW_AREA],
                                      ButtonPressMask, False,
                                      (XtEventHandler)point_in_delnode, NULL);
                    XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                    XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                    } 
                 else { /* Sensitivitaet entfernen */                    
                      XtRemoveEventHandler(Widgetliste[DRAW_AREA],
                                           ButtonPressMask, False,
                                           (XtEventHandler)point_in_delnode, NULL);
                      XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                      XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                      }
                 break;

             case NODE_INFO_BUTTON: /* Rechner veraendern */
                  if (addEvent) { /* neue Sensitivitaet */
                     XtAddEventHandler(Widgetliste[DRAW_AREA],
                                       ButtonPressMask, False,
                                       (XtEventHandler)point_in_modifynode, NULL);
                     XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False); 
                     XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                     }
                  else { /* Sensitivitaet entfernen */
                       XtRemoveEventHandler(Widgetliste[DRAW_AREA],
                                            ButtonPressMask, False,
                                            (XtEventHandler)point_in_modifynode, NULL);
                       XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True); 
                       XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                       }
                  mapScales(addEvent, touched_Function);
                  break;

             case ARC_INS_BUTTON: /* Beziehung einfuegen */
                  if (addEvent) { /* neue Sensitivitaet */
                     XtAddCallback(Widgetliste[DRAW_AREA], 
                                   XmNinputCallback, (XtCallbackProc)insert_arcCB, NULL);
                     XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                     XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                     }
                  else { /* Sensitivitaet entfernen */
                       XtRemoveCallback(Widgetliste[DRAW_AREA],
                                        XmNinputCallback, (XtCallbackProc)insert_arcCB, NULL);
                       XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                       XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                       }
                  break; 
              
             case ARC_DEL_BUTTON: /* Beziehung veraendern */
                  if (addEvent) { /* neue Sensitivitaet */
                     XtAddCallback(Widgetliste[DRAW_AREA],
                                   XmNinputCallback, (XtCallbackProc)delete_arcCB, NULL); 
                     XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                     XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                     }  
                  else { /* Sensitivitaet entfernen */
                       XtRemoveCallback(Widgetliste[DRAW_AREA],
                                        XmNinputCallback, (XtCallbackProc)delete_arcCB, NULL);
                       XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                       XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                       }
                  break;

             case ARC_INFO_BUTTON: /* Beziehung veraendern */
                  if (addEvent) { /* neue Sensitivitaet */
                     XtAddCallback(Widgetliste[DRAW_AREA],
                                   XmNinputCallback, (XtCallbackProc)modify_arcCB, NULL);
                     XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], False);
                     XtSetSensitive(Widgetliste[RUN_MENU_BUTT], False);
                     }
                  else { /* Sensitivitaet entfernen */
                       XtRemoveCallback(Widgetliste[DRAW_AREA],
                                   XmNinputCallback, (XtCallbackProc)modify_arcCB, NULL);
                       XtSetSensitive(Widgetliste[FILE_MENU_BUTTON], True);
                       XtSetSensitive(Widgetliste[RUN_MENU_BUTT], True);
                       }
                  mapScales(addEvent, touched_Function);
                  break;

             default:
                  XBell(display, 0);
                  wprint(unknown_event, "main.c : installEventHandler");
                  break;
     }          
}


void set_sensitive(exception, sensitive)
     int  exception; /* nicht zu (de)sensitivierender Knopf */
     Bool sensitive; /* de- oder sensitivieren */
/*
 * AU: Sensitivitaet der Knoepfe steuern
 */
{   
   int i;

   for (i = NODE_INS_BUTTON; i < END_FUNC_BUTTON; i++) 
       if (i != exception) XtSetSensitive(Widgetliste[i], sensitive);

   XtSetSensitive(Widgetliste[END_FUNC_BUTTON], !sensitive); 
}    



void mapScales(map, touched_Function)
     Bool map;
     int  touched_Function; /* Schieberegler eines Rechners oder
                             * einer Verbindung */
/*
 * AU: Anzeige der Schieberegler
 */
{
 int i;
 int endloop;

 switch (touched_Function)
 {
        case NODE_INFO_BUTTON: /* Rechner veraendern */
 
             endloop = NODE_QUOR_ONE + data.numb_quor;

             /* Anzeige der Schieberegler */
             for (i = NODE_QUOR_ONE; i < endloop; i++)  
                 XtVaSetValues(Widgetliste[i], 
                               XmNmappedWhenManaged, map, 
                               NULL);
 
             endloop = NODE_LABEL_ONE + data.numb_quor;

             /* Anzeige der Beschriftung */
             for (i = NODE_LABEL_ONE; i < endloop; i++) 
                 XtVaSetValues(Widgetliste[i],
                               XmNmappedWhenManaged, map, 
                               NULL);
             break;

        case ARC_INFO_BUTTON: /* Beziehung veraendern */

             endloop = ARC_LABEL_ONE + data.numb_quor;
         
              /* Anzeige der Beschriftung */
             for (i = ARC_LABEL_ONE; i < endloop; i++)
                 XtVaSetValues(Widgetliste[i],
                               XmNmappedWhenManaged, map, 
                               NULL);
          
             endloop = ARC_MOD_ONE + data.numb_quor;

             /* Anzeige der Schieberegler */                                  
             for (i = ARC_MOD_ONE; i < endloop; i++) 
                 XtVaSetValues(Widgetliste[i],
                               XmNmappedWhenManaged, map, 
                               NULL);
             break;


        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c : mapScales");
                               
             break;
 }
}


void clear_Nodelist()
/*
 * AU: Auswahlliste loeschen
 */
{
   XmListDeleteAllItems(Widgetliste[NODE_LIST]);
}



void fill_Nodelist()
/*
 * AU: Auswahlliste mit den Rechnern der Topologie und einem Eintrag
 *     fuer die Auswahl eines virtuellen Rechners fuellen 
 * ST: extractListString()
 */
{
   XmString list_string;
   NODE     *fill_node;    /* einzufuegender Rechner */
   char     name[40];      /* Bezeichnung des Eintrags */
   int      nodeCount = 1; /* Anzahl der Eintraege */

   for (fill_node  = ptrNetObjectListFront; fill_node != NULL;
                                            fill_node = fill_node->nextnode) { 
          extractListString(fill_node, name); 
          list_string = XmStringCreateLtoR(name, charset);
          XmListAddItemUnselected(Widgetliste[NODE_LIST], list_string, 1); 
          nodeCount++;
          } 

   /* Eintrag zur Auswahl eines virtuellen Rechners */
   list_string = XmStringCreateLtoR(virtual_node, charset);

   XmListAddItemUnselected(Widgetliste[NODE_LIST], list_string, 1);

   XtVaSetValues(Widgetliste[NODE_LIST],
                 XmNvisibleItemCount, nodeCount,
                 NULL); 

   XmStringFree(list_string);
}




/*ARGSUSED*/
void listCB (widget, clientDaten, listDaten)
     Widget               widget;      
     caddr_t              clientDaten; 
     XmListCallbackStruct *listDaten;
/*
 * AU: Selektion eines Eintrags in der Auswahlliste
 * ST: setNodeListSensitive()
 *     installEventHandler()
 */
{
   char *first_blank;
   char *cstring;
   char s[43];

   XmStringGetLtoR(listDaten->item, charset, &cstring); 

   strncpy(s, cstring, 40);

   if (s[0] != STAR) {  
      setNodeListSensitive(False); /* keine weitere Auswahl moeglich */
      strncpy(data.title, cstring, 40);
      strcpy(s, data.title);  

      if (strncmp(data.title, virtual_node, 17) == 0) 

          /* Eintrag fuer einen virtuellen Rechner */
          XtManageChild(Widgetliste[NODE_NAME_PROMPT]);
      else {
           first_blank = (char *)strchr(s, BLANK);

           if (first_blank == NULL) { /* kein Eigennamen vergeben */
              strncpy(data.identity,  data.title, 40);
              strcpy(data.name,  SPACE);
              }
           else { /* Eintrag besitzt einen selbst vergebenen Namen */
                *first_blank = EOS;
                strncpy(data.identity, s, 40);
                first_blank++;
                strncpy(data.name, first_blank, 40);
                }

           data.node_version = 0; 
           data.virtualNode = False;
           installEventHandler(NODE_INS_BUTTON, True); /* Arbeitsflaeche
                                                        * sensitivieren */
           }  
      }
   else {
        XBell(display, 0);
        wprint(nodealreadyinsert);
        }
}



void setNodeListSensitive(sensitive)
     Bool sensitive; /* de- oder sensitivieren */
/*
 * AU: Auswahlliste de- und sensitivieren
 */
{
   XtSetSensitive(Widgetliste[NODE_LIST], sensitive);
}



/*ARGSUSED*/
void start_rubber_band (widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechnerumriss am Ausgangspunkt zeichnen
 * ST: drawRubberArc()
 *     drawRubberRectangle()
 */
{    
   data.last_x = data.start_x = eventAdr->xcrossing.x;
   data.last_y = data.start_y = eventAdr->xcrossing.y;
      
   if (data.virtualNode) drawRubberArc(data.start_x, data.start_y);
   else                  drawRubberRectangle(data.start_x, data.start_y);
}


/*ARGSUSED*/
void interrupt_rubber_band (widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechnerumriss beim Verlassen des Arbeitsbereichs zeichnen
 * ST: drawRubberArc()
 *     drawRubberRectangle()
 */
{
   /* Alten Umriss loeschen */
   if (data.virtualNode) drawRubberArc(data.start_x, data.start_y);
   else                  drawRubberRectangle(data.start_x, data.start_y);
}


/*ARGSUSED*/
void track_rubber_band (widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechnerumriss beim Verschieben zeichnen
 * ST: drawRubberArc()
 *     drawRubberRectangle()
 */
{    
   /* Alten Umriss loeschen */ 
   if (data.virtualNode) drawRubberArc(data.start_x, data.start_y);
   else                  drawRubberRectangle(data.start_x, data.start_y);

   data.last_x = eventAdr->xmotion.x;
   data.last_y = eventAdr->xmotion.y;

   if (data.virtualNode) drawRubberArc(data.last_x, data.last_y);
   else                  drawRubberRectangle(data.last_x, data.last_y);

   data.start_x = data.last_x;
   data.start_y = data.last_y;   
}
              
  
/*ARGSUSED*/
void end_nodeinsert_rubber_band(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Positionierung eines einzufuegenden Rechners
 * ST: drawRubberArc()
 *     drawRubberRectangle()
 *     installEventHandler()
 *     validPosition()
 *     markListPos()
 *     add_GraphicObject()
 *     addRepresentatives()
 *     change_GraphicObject()
 *     drawStringsTerminal() 
 */
{
   char     helpstring[40];   /* Bezeichnung */
   char     stringVersion[3]; /* Versionsnummer */
        
   /* Alten Umriss loeschen */ 
   if (data.virtualNode) drawRubberArc(data.last_x, data.last_y);
   else                  drawRubberRectangle(data.last_x, data.last_y);

   /* Sensitivitaet entfernen */
   installEventHandler(NODE_INS_BUTTON, False);
          
   if (validPosition(data.last_x, data.last_y, NULL)) {

      /* gueltige Position */
      if (!data.virtualNode) {
         markListPos(data.title); /* Eintrag als ausgewaehlt
                                   * markieren */
         strcpy(helpstring, data.identity);
         data.node_ident = NULL;
         }
      else { /* virtueller Rechner */
           sprintf(stringVersion, "%d", data.node_version);
           strcpy(helpstring, data.identity);
           strcat(helpstring, "[");
           strcat(helpstring, stringVersion);
           strcat(helpstring, "]");
           data.virtualNode = False;
           strcpy(data.name, SPACE);
           strcpy(data.identity, helpstring);
    
           XmListDeselectPos(Widgetliste[NODE_LIST], 1);
 
           }

     /* Rechner der Voting-Struktur hinzufuegen */
     add_GraphicObject(data.name, helpstring, data.node_version);

     /* Liste der Repraesentanten hinzufuegen */
     addRepresentatives(ptrGraphicObjectList, data.node_ident);

     /* Rechnerkoordinaten belegen */
     change_GraphicObject(data.last_x, data.last_y, ptrGraphicObjectList);
      
     /* Rechnersymbol und Beschriftungen zeichnen */    
     drawStringsTerminal(data.last_x, data.last_y, data.name, data.identity,
                         data.node_vote, data.node_quor, data.node_version);
     }
   else { /* ungueltige Position */
        XmListDeselectAllItems(Widgetliste[NODE_LIST]);
        XBell(display, 0);
        wprint(nonvalid_position);
        }

}


Bool markListobject(listobject)
     GRAPHIC_NODE *listobject; /* Rechner der Voting-Struktur */
/*
 * AU: Suche einen Rechnereintrag und ersetze ihn durch den
 *     "Bereits eingetragen"-Text
 */
{
   XmString matchstring, new_item_string;
   char itemstring[40], newitemstring[41];
   char *stern = LIST_MARK;
   int  *pos_list;
   int  pos_cnt;

   if (listobject->version == 0) { /* kein virtueller Rechner */
      strcpy(itemstring, listobject->ident);
      strcat(itemstring, SPACE);
      strcat(itemstring, listobject->name);

      matchstring = XmStringCreateLtoR(itemstring, charset);

      if (XmListGetMatchPos(Widgetliste[NODE_LIST], matchstring,
                            &pos_list, &pos_cnt)) {
          strcpy(newitemstring, stern);
          strcat(newitemstring, itemstring);
          new_item_string = XmStringCreateLtoR(newitemstring, charset);
          XmListReplaceItemsPos(Widgetliste[NODE_LIST], &new_item_string,
                               pos_cnt, pos_list[0]);
          XmListDeselectPos(Widgetliste[NODE_LIST], pos_list[0]);
          XtFree((char *)pos_list);
          XmStringFree(new_item_string);
          }

      XmStringFree(matchstring);
      return True;
      }

   return False;                                  
}


Bool markListPos(title) 
     char title[]; /* Rechnerbezeichnung */
/*
 * AU: Ersetze den selektierten Eintrag durch den "Bereits eingetragen"-
 *     Text
 */
{ 
   char     *stern = LIST_MARK;
   char     helpstring[40];
   int      *poslist;
   int      poscnt;                  
   XmString motif_string;

   strcpy(helpstring, stern);
   strcat(helpstring, title);
 
   motif_string = XmStringCreateLtoR(helpstring, charset);
   
   if (XmListGetSelectedPos(Widgetliste[NODE_LIST], &poslist, &poscnt)) {
      XmListReplaceItemsPos(Widgetliste[NODE_LIST], &motif_string,
                            poscnt, poslist[0]);
      XmListDeselectPos(Widgetliste[NODE_LIST], poslist[0]);
      XtFree((char *)poslist);
      XmStringFree(motif_string);
      /* wird auf Stack erzeugt (???): XtFree(stern); */
      /* XtFree(helpstring); */
      return True;
      }

   XmStringFree(motif_string);
   /* XtFree(stern); */
   /* XtFree(helpstring); */

   return False;
}


/*ARGSUSED*/
void point_in_delnode(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechner der Voting-Struktur loeschen
 * ST: remove_arcs()
 *     remove_ident()
 *     EditCB()
 *     node_hit()
 *     restore_ListObject()
 *     clearArea()
 *     redrawArea()
 */
{
   int param;
   GRAPHIC_NODE *ptrDelNode = NULL;

   ptrDelNode = /* selektierter Rechner */
         (GRAPHIC_NODE *)node_hit(eventAdr->xbutton.x, eventAdr->xbutton.y);

   if (restore_Listobject(ptrDelNode)) { /* Eintrag Auswahlliste aktualisiert */

       if ((ptrDelNode->previous == NULL) && (ptrDelNode->next == NULL)) {

           /* letzter Rechner der Voting-Struktur geloescht */
           remove_arcs(ptrDelNode);     /* Verbindungen entfernen */
           remove_ident(ptrDelNode);    /* Liste der Repraesentanten loeschen */
           ptrGraphicObjectList = NULL; /* einziges Listenelement loeschen */
           param = END_FUNC_BUTTON;
           EditCB(NULL, (caddr_t)&param, NULL);  /* Funktion beenden */
           }
       else {
            if (ptrDelNode->previous == NULL) { /* erstes Element */
               remove_arcs(ptrDelNode); /* Verbindungen entfernen */
               remove_ident(ptrDelNode); /* Liste der Repraesentanten loeschen */
               ptrGraphicObjectList = ptrDelNode->next;   
               ptrDelNode->next->previous = NULL;
               }
            else {
                 if (ptrDelNode->next == NULL) { /* letztes Element */
                    remove_arcs(ptrDelNode);  /* Verbindungen entfernen */
                    remove_ident(ptrDelNode); /* Liste der Repraesentanten loeschen */
                    ptrDelNode->previous->next = NULL;
                    }
                 else { /* inneres Element */
                      remove_arcs(ptrDelNode);  /* Verbindungen entfernen */
                      remove_ident(ptrDelNode); /* Liste der Repraesentanten loeschen */
                      ptrDelNode->previous->next = ptrDelNode->next;
                      ptrDelNode->next->previous = ptrDelNode->previous;
                      }
                 } 
            }
      XtFree((char *)ptrDelNode);

      clearArea();

      redrawArea(ptrGraphicObjectList);
      }
}


Bool restore_Listobject(listobject)
     GRAPHIC_NODE *listobject; /* wieder selektierbarer Eintrag */
/*
 * AU: Rechner der Auswahlliste als wieder selektierbar kennzeichnen 
 * NB: Rueckgabewert False bei Fehler, sonst True
 */
{
   XmString matchstring, new_item_string;
   char itemstring[40];
   char *stern = LIST_MARK;
   char *newstring;
   int  *pos_list;
   int  pos_cnt; 


   if (listobject != NULL) {

      if (listobject->version == 0) { /* physikalischer Rechner */

          strcpy(itemstring, stern);
          strcat(itemstring, listobject->ident);
          strcat(itemstring, SPACE);
          strcat(itemstring, listobject->name);
   
          matchstring = XmStringCreateLtoR(itemstring, charset);
  
          /* Suche nach dem passenden Eintrag */
          if (XmListGetMatchPos(Widgetliste[NODE_LIST], matchstring,
                                &pos_list, &pos_cnt)) {
               
             newstring = (char *)strstr(itemstring, "R");
             new_item_string = XmStringCreateLtoR(newstring, charset);
 
             XmListReplaceItemsPos(Widgetliste[NODE_LIST], &new_item_string,
                                   pos_cnt, pos_list[0]);
             XmListDeselectPos(Widgetliste[NODE_LIST], pos_list[0]);
             wprint(Nodedeleted, newstring);

             XtFree((char *)pos_list);
             XmStringFree(new_item_string);
             }
          else {
               XBell(display, 0);
               wprint(missingListEntry);
               return False;
               }

          XmStringFree(matchstring);
         }
       else wprint(virtNodedeleted);
      }
   else return False;

   /* XtFree(itemstring); */
   /* XtFree(stern); */

   return True;
}


/*ARGSUSED*/
void point_in_movenode(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; 
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechner der Voting-Struktur verschieben
 * ST: node_hit()
 *     drawRubberArc()
 *     drawRubberRectangle()
 */
{   
   GRAPHIC_NODE *ptrMoveNode = NULL;

   data.last_x = data.start_x = eventAdr->xbutton.x;
   data.last_y = data.start_y = eventAdr->xbutton.y;

   ptrMoveNode = /* selektierter Rechner */
         (GRAPHIC_NODE *)node_hit(eventAdr->xbutton.x, eventAdr->xbutton.y);

   if (ptrMoveNode != NULL) { /* gueltige Selektion */

      if (ptrMoveNode->version != 0) { /* virtueller Rechner */
         data.virtualNode = True;
         drawRubberArc(data.start_x, data.start_y);
         }
      else drawRubberRectangle(data.start_x, data.start_y);

      /* Arbeitsbereich sensitivieren */
      XtAddEventHandler(Widgetliste[DRAW_AREA],ButtonMotionMask, False,
                        (XtEventHandler)track_rubber_band, NULL);
      XtAddEventHandler(Widgetliste[DRAW_AREA], LeaveWindowMask, False,
                        (XtEventHandler)leave_window_node, (caddr_t)ptrMoveNode);
      XtAddEventHandler(Widgetliste[DRAW_AREA], ButtonReleaseMask, False,
                        (XtEventHandler)end_move_node, (caddr_t)ptrMoveNode);
      }
}


/*ARGSUSED*/
void point_in_modifynode(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Wertebelegung Rechner veraendern
 * ST: node_hit()
 */
{
   int          i;
   char         node_label[40];
   XmString     motif_label;
   GRAPHIC_NODE *ptrModifyNode;

   ptrModifyNode = /* selektierter Rechner */
         (GRAPHIC_NODE *)node_hit(eventAdr->xbutton.x, eventAdr->xbutton.y);

   if (ptrModifyNode != NULL) { /* gueltige Selektion */
      XtVaSetValues(Widgetliste[NODE_MODIFY_CONT], 
                    XmNuserData, ptrModifyNode, 
                    NULL);

      /* Schieberegler des Dialogfensters mit gegenwaertigen Werten belegen */
      XtPopup(Widgetliste[NODE_MODIFY_SHELL], XtGrabExclusive);  
      data.node_vote = ptrModifyNode->vote;

      for (i = 0; i < 5; i++) data.node_quor[i] = ptrModifyNode->quor[i]; 

      XmScaleSetValue(Widgetliste[NODE_VOTE_SIX], ptrModifyNode->vote);
      /* falls Werte in ptrModifyNode->quor[x] nicht initialisiert sind und ungueltige
       * Slider-Werte aufweisen, gibt es SEGFAULTS in XtAppError(), falls die
       * Intrinsics-Version von XtError() benutzt wird; siehe auch XtError() in main.c
       */
      XmScaleSetValue(Widgetliste[NODE_QUOR_ONE], ptrModifyNode->quor[0]);
      XmScaleSetValue(Widgetliste[NODE_QUOR_TWO], ptrModifyNode->quor[1]);
      XmScaleSetValue(Widgetliste[NODE_QUOR_THREE], ptrModifyNode->quor[2]);
      XmScaleSetValue(Widgetliste[NODE_QUOR_FOUR], ptrModifyNode->quor[3]);
      XmScaleSetValue(Widgetliste[NODE_QUOR_FIVE], ptrModifyNode->quor[4]);

      /* Kopfzeile bilden */
      strcpy(node_label, ptrModifyNode->ident);
      strcat(node_label, SPACE);
      strcat(node_label, ptrModifyNode->name);
      motif_label = XmStringCreateLtoR(node_label, charset);

      XtVaSetValues(Widgetliste[NODE_MODIFY_NAME],
                    XmNlabelString, motif_label, 
                    NULL);

      XmStringFree(motif_label);    
      }
}

 
/*ARGSUSED*/
void leave_window_node(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; 
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechnerumriss beim Verlassen des Arbeitsbereichs zeichnen
 * ST: drawRubber Arc()
 *     drawRubberRectangle()
 */
{
   GRAPHIC_NODE *ptrMoveNode;

   ptrMoveNode = (GRAPHIC_NODE *)clientDaten;

   /* Sensitivitaet des Arbeitsbereichs entfernen */
   XtRemoveEventHandler(Widgetliste[DRAW_AREA],ButtonMotionMask, False,
                        (XtEventHandler)track_rubber_band, NULL);
   XtRemoveEventHandler(Widgetliste[DRAW_AREA], ButtonReleaseMask, False,
                        (XtEventHandler)end_move_node, (caddr_t)ptrMoveNode);
   XtRemoveEventHandler(Widgetliste[DRAW_AREA], LeaveWindowMask, False,
                        (XtEventHandler)leave_window_node, (caddr_t)ptrMoveNode);

   if (data.virtualNode) drawRubberArc(data.last_x, data.last_y);
   else                  drawRubberRectangle(data.last_x, data.last_y);

   data.virtualNode = False;
}


/*ARGSUSED*/
void end_move_node(widget, clientDaten, eventAdr, weiterArb)
     Widget       widget;      /* den Event-Hdl. aufrufendes Widget */
     caddr_t      clientDaten; /* hier nicht verwendet */
     XEvent       *eventAdr;   /* Zeiger auf ausloesendes X-Event */
     Boolean      *weiterArb;  /* weitere Event-Hdl. verarbeiten? */
/*
 * AU: Rechnerumriss am Ende des Verschiebens zeichnen
 * ST: change_GraphicObject()
 *     clearArea()
 *     redrawArea() 
 */
{
   GRAPHIC_NODE *ptrMoveNode;

   ptrMoveNode = (GRAPHIC_NODE *)clientDaten;

   if (validPosition(eventAdr->xbutton.x, eventAdr->xbutton.y, ptrMoveNode))
       /* falls eine gueltige Postion gewaehlt wurde, neue Koordinaten vermerken */ 
       change_GraphicObject(eventAdr->xbutton.x, eventAdr->xbutton.y,
                            ptrMoveNode);
   else { /* Fehler */
        XBell(display, 0);
        wprint(nonvalid_position);
        }

   /* Sensitivitaet des Arbeitsbereichs entfernen */
   XtRemoveEventHandler(Widgetliste[DRAW_AREA],ButtonMotionMask, False,
                        (XtEventHandler)track_rubber_band, NULL);   
   XtRemoveEventHandler(Widgetliste[DRAW_AREA], ButtonReleaseMask, False,
                        (XtEventHandler)end_move_node, (caddr_t)ptrMoveNode);
   XtRemoveEventHandler(Widgetliste[DRAW_AREA], LeaveWindowMask, False,
                        (XtEventHandler)leave_window_node, (caddr_t)ptrMoveNode);


   /* Neue Darstellung abbilden */
   clearArea();
           
   redrawArea(ptrGraphicObjectList);

   data.virtualNode = False; 
}


/*ARGSUSED*/
void delete_arcCB(widget, client_data, call_data)
     Widget                      widget;
     caddr_t                     client_data;
     XmDrawingAreaCallbackStruct *call_data;
/*
 * AU: Verbindung zwischen Rechnern loeschen
 * ST: node_hit()
 *     deleteArc()
 *     clearArea()
 *     redrawArea()
 */
{
 static GRAPHIC_NODE *first_marked_node, *next_marked_node;
 XEvent *event = call_data->event;

 switch (event->xany.type) 
 {
        case ButtonPress: 
             first_marked_node = /* 1. Rechner selektieren */
                  (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);
             break;

        case ButtonRelease:
             next_marked_node = /* 2.Rechner selektieren */
                 (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);
         
             /* Zwei verschiedene Rechner wurden selektiert ? */
             if ((first_marked_node != NULL) && (next_marked_node != NULL) && 
                 (first_marked_node != next_marked_node)) {

                 if (deleteArc(first_marked_node, next_marked_node)) { 
                    wprint(goodarcdelete, 
                       first_marked_node->ident, next_marked_node->ident);
                    clearArea();
                    redrawArea(ptrGraphicObjectList);
                    }
                 else {
                      XBell(display, 0);
                      wprint(badarcdelete);
                      }

                 first_marked_node = next_marked_node = NULL; 
                 }
             break;

        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c : delete_arcCB");
             break;
 }
}


/*ARGSUSED*/
void modify_arcCB(widget, client_data, call_data)
     Widget   widget;
     caddr_t  client_data;
     XmDrawingAreaCallbackStruct  *call_data;
/*
 * AU: Wertebelegung einer Verbindung veraendern
 * ST: node_hit()
 *     arc_already_exists()
 */
{
 static   GRAPHIC_NODE *starting_node, *end_node; /* Beteiligte Rechner */
          GRAPHIC_ARC  *ptrModifyArc;             /* Verbindung */
          XEvent       *event = call_data->event;
          char         arc_label[100]; /* Titelzeile */
          XmString     motif_label;
          int          i;

 switch (event->xany.type)
 {
        case ButtonPress:
             starting_node = /* 1. Rechner selektieren */
                 (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);
             break;

        case ButtonRelease:
             end_node = /* 2. Rechner selektieren */
                 (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);

             /* Zwei verschiedene Rechner wurden selektiert ? */
             if ((starting_node != NULL) && (end_node != NULL) && 
                 (starting_node != end_node)) {

                /* Verbindung existiert ? */
                ptrModifyArc = (GRAPHIC_ARC *)
                         arc_already_exists(starting_node, end_node); 

                if (ptrModifyArc != NULL) {
                   data.starting_node = starting_node;
                   data.end_node = end_node;

                   /* Aufbau Titelzeile */
                   strncpy(arc_label, modifyarc, 40);
                   strncat(arc_label, starting_node->ident, 30);
                   strcat(arc_label, SPACE);
                   strncat(arc_label, starting_node->name, 30);
                   strcat(arc_label, " und ");
                   strncat(arc_label, end_node->ident, 30);
                   strcat(arc_label, SPACE);
                   strncat(arc_label, end_node->name, 30);
                   motif_label = XmStringCreateLtoR(arc_label, charset);

                   XtVaSetValues(Widgetliste[ARC_MODIFY_NAME],
                                 XmNlabelString, motif_label, 
                                 NULL);

                   /* Schieberegler des Dialogfensters mit den aktuellen
                    * Werten belegen */
                   for (i = 0; i < 5; i++) 
                       data.arc_value[i] = ptrModifyArc->arc_value[i];
     
                   XmScaleSetValue(Widgetliste[ARC_MOD_ONE], 
                                   data.arc_value[0]);
                   XmScaleSetValue(Widgetliste[ARC_MOD_TWO], 
                                   data.arc_value[1]);
                   XmScaleSetValue(Widgetliste[ARC_MOD_THREE], 
                                   data.arc_value[2]);
                   XmScaleSetValue(Widgetliste[ARC_MOD_FOUR], 
                                   data.arc_value[3]);
                   XmScaleSetValue(Widgetliste[ARC_MOD_FIVE], 
                                   data.arc_value[4]); 
              
                   /* XtFree(arc_label); */
                   XmStringFree(motif_label);

                   XtPopup(Widgetliste[ARC_MODIFY_SHELL], XtGrabExclusive);
                   }
                else {
                     XBell(display, 0);
                     wprint(noConnection);
                     }    
                }
             starting_node = end_node = NULL;
             break;

        default:  
             XBell(display, 0);
             wprint(unknown_event, "main.c : modify_arcCB");
             break;
 } 
}


/*ARGSUSED*/
void insert_arcCB(widget, client_data, call_data)
     Widget                      widget;
     caddr_t                     client_data;
     XmDrawingAreaCallbackStruct *call_data;
/* 
 * AU: Verbindung zwischen Rechnern einfuegen
 * ST: node_hit()
 *     redrawArea()
 *     setGraphicArcWeight()
 *     insertArc()
 */
{
 static GRAPHIC_NODE *starting_node, *end_node; /* beteiligte Rechner */
 XEvent *event =     call_data->event;
 int    i; 

 switch (event->xany.type)
 {
        case ButtonPress:
             starting_node = /* 1. Rechner selektieren */
                 (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);
             break;

        case ButtonRelease:
              end_node = /* 2. Rechner selektieren */
                 (GRAPHIC_NODE *)node_hit(event->xbutton.x, event->xbutton.y);

              /* Zwei verschieden Rechner wurden selektiert */
              if ((starting_node != NULL) && (end_node != NULL) && 
                  (starting_node != end_node)) {

                 if (insertArc(starting_node, end_node)) { /* Verbindung einfuegen */
                    for (i = 0; i < 5; i++)
                        setGraphicArcWeight(starting_node, end_node, i, 0); 
                    redrawArea(ptrGraphicObjectList);
                    wprint(insertarc, starting_node->ident, end_node->ident);
                   }
                 else {
                      XBell(display, 0);
                      wprint(arcalreadyinsert);
                      } 
                 }

             starting_node = end_node = NULL;
             break;
        
        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c : insert_arcCB");
             break;
 }
}

/*ARGSUSED*/        
void RunCB (widget, clientDaten, aufrufDaten)
     Widget  widget;
     caddr_t clientDaten, aufrufDaten;
/*
 * AU: Betaetigung der Menuepunkte des Analysemenues verarbeiten
 * ST: checkVQM();
 *     computeVQM();
 *     computeReli();
 */
{
   static GRAPHIC_NODE *ptrGraphicObjectRoot; /* Wurzel der Voting-Struktur */
   int                 i, position;           /* Laufvariablen */
   XmTextPosition      lastPos;               /* Textende */
   char                stringReli[255];       /* Wahrscheinlichkeit*/

   position = RESULTS_ONE + data.numb_quor;   /* Anzeige der Ergebnisse steueren */
 		                   
   switch (*((int*)clientDaten))
   {
	  case CHECK_VQM_BUTT : /* Voting-Struktur pruefen */
	       ptrGraphicObjectRoot =
			 (GRAPHIC_NODE *)checkVQM(ptrGraphicObjectList);
	       if (ptrGraphicObjectRoot != NULL) { /* fehlerhafte Voting-Struktur */
		   wprint(validVQM, ptrGraphicObjectRoot->ident);
		   XtSetSensitive(Widgetliste[COMPUTE_VQM_BUTT], True);
		  }
							
               break;
					
          case COMPUTE_VQM_BUTT : /* Quorenliste berechnen */
               /* print_Graph(); */
               /* print_GraphicObjects(); */
               if (computeVQM(ptrGraphicObjectRoot)) { /* erfolgreiche Berechnung */
	          XtSetSensitive(Widgetliste[COMPUTE_VQM_BUTT], False);
	          XtSetSensitive(Widgetliste[COMPUTE_RELI_BUTT], True);
		 }
               else wprint(invalidStructure, ptrCurrRefList->ptrRefItem->ident);
	
	       break;
	
          case COMPUTE_RELI_BUTT : /* Verfuegbarkeit der Operationen berechnen */

	       for (i = 0; i < data.numb_quor; i++) { /* fuer jede Operation */
		   data.reliability[i] = 0; 

		   computeReli(ptrNetObjectListFront, i); /* Berechnung */

		   XtSetSensitive(Widgetliste[COMPUTE_RELI_BUTT], False);
           
                   /* Bezeichnung der Netztopologiedatei anzeigen */
                   XmTextSetString(Widgetliste[RESULT_TOPFILE], 
                                   data.filename_TOP);
                   lastPos = XmTextGetLastPosition(Widgetliste[RESULT_TOPFILE]);
                   XmTextShowPosition(Widgetliste[RESULT_TOPFILE], lastPos);

                   /* Bezeichnung der Voting-Struktur-Datei anzeigen */
                   XmTextSetString(Widgetliste[RESULT_VOTFILE], 
                                   data.filename_VOT);
                   lastPos = XmTextGetLastPosition(Widgetliste[RESULT_VOTFILE]);
                   XmTextShowPosition(Widgetliste[RESULT_VOTFILE], lastPos);

                   /* Ergebniss in Zeichen umwandeln */
                   sprintf(stringReli, "%f", data.reliability[i]);

                   /* Anzeige in Abhaengigkeit der Operationenanzahl */
                   switch (i + RESULT_SERVICE1)
                   {
                          case RESULT_SERVICE1 :                 
                               XmTextSetString(Widgetliste[RESULT_SERVICE1],
                                               stringReli);
                               break;

                          case RESULT_SERVICE2 :
                               XmTextSetString(Widgetliste[RESULT_SERVICE2],
                                               stringReli);
                               break;

                          case RESULT_SERVICE3 :
                               XmTextSetString(Widgetliste[RESULT_SERVICE3],
                                               stringReli);
                               break;

                          case RESULT_SERVICE4 :
                               XmTextSetString(Widgetliste[RESULT_SERVICE4], 
                                               stringReli); 
                               break; 

                          case RESULT_SERVICE5 :
                               XmTextSetString(Widgetliste[RESULT_SERVICE5], 
                                               stringReli); 
                               break; 

                          default:
                               XBell(display, 0);
                               wprint(unknown_event, "main.c : part 1 RunCB");
                               
                               break;

                   }
                   }
               /* Anzeige veranlassen */
               for (i = RESULTS_ONE; i < position; i++) {
                   XtVaSetValues(Widgetliste[i],
                                 XmNmappedWhenManaged, True,
                                 NULL);
                   }

               /* andere Operationen verbergen */
               for (i = position; i < RESULTS_FIVE + 1; i++) {
                   XtVaSetValues(Widgetliste[i],
                                 XmNmappedWhenManaged, False,
                                 NULL);
                   }

               /* Anzeige melden und veranlassen */
               XtPopup(Widgetliste[PRES_RES_SHELL], XtGrabNone);
               XBell(display, 0);
	       break;

          default:
	       XBell(display, 0);
	       wprint(unknown_event, "main.c : part 2 RunCB");
	
	       break;
    }
}



/*ARGSUSED*/
void modifyCB(widget, clientDaten, aufrufDaten)
     Widget  widget;
     caddr_t clientDaten, aufrufDaten;
/*
 * AU: Neue Wertebelegungen fuer Rechner und Verbindungen uebernehmen
 * ST: add_GraphicVoteQuor()
 *     clearArea()
 *     redrawArea()
 *     setGraphicArcWeight()
 */
{
   int          i;
   GRAPHIC_NODE *ptrModifyNode;

   switch (*((int*)clientDaten))
   {
          case NODE_MODIFY_OK_BUTT : /* Wertebelgung Rechner */
               XtVaGetValues(Widgetliste[NODE_MODIFY_CONT],
                             XmNuserData, &ptrModifyNode, 
                             NULL);
               add_GraphicVoteQuor(ptrModifyNode, data.node_vote, 
                                   data.node_quor[0],
                                   data.node_quor[1],
                                   data.node_quor[2],
                                   data.node_quor[3],
                                   data.node_quor[4]);

               data.node_vote = 0;
           
               XtPopdown(Widgetliste[NODE_MODIFY_SHELL]);
               
               clearArea();
               redrawArea(ptrGraphicObjectList);
               break; 

          case ARC_MODIFY_OK_BUTT : /* Wertebelegung Verbindung */
               for (i = 0; i < 5; i++)
                   setGraphicArcWeight(data.starting_node, data.end_node,
                                      i, data.arc_value[i]);
             
               XtPopdown(Widgetliste[ARC_MODIFY_SHELL]);

               data.starting_node = data.end_node = NULL;
               clearArea();
               redrawArea(ptrGraphicObjectList);
               break;

          case NODE_MODIFY_CANCEL_BUTT : /* Abbruch Operation f. Rechner */
               data.node_vote = 0;
     
               XtPopdown(Widgetliste[NODE_MODIFY_SHELL]);  
               break;
       
          case ARC_MODIFY_CANCEL_BUTT : /* Abbruch Operation f. Verbindungen */
               XtPopdown(Widgetliste[ARC_MODIFY_SHELL]);
               break;

          default:
               XBell(display, 0);
               wprint(unknown_event, "main.c : modifyCB");
               break;
   } 

   /* Initialisierung der globalen Variablen */
   for (i = 0; i < 5; i++) {
       data.node_quor[i] = 0;
       data.arc_value[i] = 0;
       }
}


/*ARGSUSED*/
void scaleCB(widget, clientDaten, Scaleinfos)
     Widget                widget;
     caddr_t               clientDaten;
     XmScaleCallbackStruct *Scaleinfos;
/*
 * AU: Schieberegler zur Veraenderung der Wertebelegungen ablesen
 * NB: Jede Veraenderung der Schieberegler wird registriert. Bis zur
 * Entscheidung, ob die Werte zu uebernehmen sind werten die Resultate
 * in den globalen Variablen festgehalten
 */
{
   if (Scaleinfos->reason == XmCR_VALUE_CHANGED) {

      switch (*((int*)clientDaten))
      {
          /* Schieberegler Rechner */ 
          case NODE_QUOR_ONE :
               XmScaleGetValue(Widgetliste[NODE_QUOR_ONE], 
                               &data.node_quor[0]);  
               break;

          case NODE_QUOR_TWO :
               XmScaleGetValue(Widgetliste[NODE_QUOR_TWO], 
                               &data.node_quor[1]);
               break;
      
          case NODE_QUOR_THREE :
               XmScaleGetValue(Widgetliste[NODE_QUOR_THREE],
                               &data.node_quor[2]);
               break;

          case NODE_QUOR_FOUR :
               XmScaleGetValue(Widgetliste[NODE_QUOR_FOUR],
                               &data.node_quor[3]);
               break;

          case NODE_QUOR_FIVE :
               XmScaleGetValue(Widgetliste[NODE_QUOR_FIVE],
                               &data.node_quor[4]);
               break;

          case NODE_VOTE_SIX :
               XmScaleGetValue(Widgetliste[NODE_VOTE_SIX], &data.node_vote);
               break;

          /* Schieberegler Verbindung */
          case ARC_MOD_ONE :
               XmScaleGetValue(Widgetliste[ARC_MOD_ONE], &data.arc_value[0]);
               break;

          case ARC_MOD_TWO :
               XmScaleGetValue(Widgetliste[ARC_MOD_TWO], &data.arc_value[1]);
               break;

          case ARC_MOD_THREE :
               XmScaleGetValue(Widgetliste[ARC_MOD_THREE], &data.arc_value[2]);
               break;

          case ARC_MOD_FOUR :
               XmScaleGetValue(Widgetliste[ARC_MOD_FOUR], &data.arc_value[3]);
               break;

          case ARC_MOD_FIVE :
               XmScaleGetValue(Widgetliste[ARC_MOD_FIVE], &data.arc_value[4]);
               break;

          default:
               XBell(display, 0);
               wprint(unknown_event, "main.c : scaleCB");
               break;
      }
      }
}



void FetchWorkingDirectory()
/*
 * AU: Pfad des gegenwaertigen Arbeitsverzeichnis und Name des
 *     ausfuehrenden Rechners ermitteln
 */
{
   char *ptrChr;

   ptrChr = getenv("HOSTNAME");

   strcpy(data.host, ptrChr);

   getwd(data.pathname); /* Globale Variable belegen */
   wprint(work_dir, data.pathname);
} 



/*ARGSUSED*/
void fileNameDialogCB(widget, client_data, call_data)
     Widget  widget;
     caddr_t client_data, call_data;
/*
 * AU: Eingabe eines neuen Dateinamens verarbeiten
 * ST: clear_GraphicObjects()
 *     file_exists()
 *     clearArea()
 *     redrawArea()
 *     clear_Nodelist()
 *     fill_Nodelist()
 */
{
 char    *name; /* Dateiname */

 switch (*((int*)client_data))
 {
        case FILE_NAME_OK_BUTT : /* Uebernahme des Dateinamens erwuenscht */ 

             if (ptrNetObjectListFront == NULL) { /* keine zugehoerige 
                                                   * Netztopologie vorhanden */
                 XBell(display, 0);  
                 wprint(missingList);
                 break;   
                 }

             if ((name =  XmTextGetString(Widgetliste[FILE_NAME_TEXT]))) {

                if (!*name || !name) { /* keine Eingabe durch Benutzer */
                   XBell(display, 0);
                   wprint(missingFileEntry);
                   }
                else {
                     if (strstr(name,".") || file_exists(name)) {
                        /* Eingabe besitzt eine Endung oder existiert bereits */
                        XBell(display, 0); 
                        wprint(invalid_filename);
                        }
                     else {
                          wprint(chosen_filename, name);
                          /* Globale Variablen mit Dateinamen belegen */
                          strncpy(data.filename_VOT, name, 254);
                          strncpy(data.filename_DRW, name, 254);

                          /* Gewaehlte Quorenanzahl global ablegen */
                          XmScaleGetValue(Widgetliste[FILE_NAME_SCALE], 
                                          &data.numb_quor);
                          XtMapWidget(Widgetliste[EDIT_BUTTONS]);
                          XtSetSensitive(Widgetliste[OPEN_TREE_BUTTON], False);
                          XtSetSensitive(Widgetliste[NEW_TREE_BUTTON], False);
                          XtSetSensitive(Widgetliste[SAVE_YN_BUTTON], True);
                          XtSetSensitive(Widgetliste[CHECK_VQM_BUTT], True);
                          clear_GraphicObjects();

                          if (zoomOpportunity) {
                             zoomOption = ZOOM_100;
                             XtSetSensitive(Widgetliste[ZOOM_100_BUTTON], False);
                             XtSetSensitive(Widgetliste[ZOOM_75_BUTTON], True);
                             XtSetSensitive(Widgetliste[ZOOM_50_BUTTON], True);
			     }

                          clearArea(); 
                          redrawArea(ptrGraphicObjectList);
                          clear_Nodelist();
                          fill_Nodelist();
                          XtPopdown(Widgetliste[FILE_NAME_SHELL]);
                          }   
                     }   
                XtFree(name);
                }
             break;

        case FILE_NAME_CANCEL_BUTT : /* Abbruch der Eingabe */

             XtSetSensitive(Widgetliste[NEW_TREE_BUTTON], True);
             XtSetSensitive(Widgetliste[OPEN_NET_BUTTON], True);
             XtSetSensitive(Widgetliste[OPEN_TREE_BUTTON], True);
             XtPopdown(Widgetliste[FILE_NAME_SHELL]);
             break;
               
        default:
             XBell(display, 0);
             wprint(unknown_event, "main.c : fileNameDialogCB");
             break;
 }
}


/*ARGSUSED*/
void store_freeGraphicCB(widget, client_data, messageinfos)
     Widget              widget;
     caddr_t             client_data;
     XmAnyCallbackStruct *messageinfos;
/*
 * AU: Sicherung einer Datei
 * ST: file_exists()
 *     getDrawingsize()
 *     saveGraphicObjects()
 *     clear_GraphicObjects()
 *     clearArea()
 *     redrawArea()
 */
{
  static Dimension xsize, ysize;     /* Dimensionen der Arbeitsflaeche */ 
  static Widget    speichernAlsText = NULL; /* Widget-ID des Textfensters */
  static Bool      warnedOfDataLoss; /* Hinweis auf moeglichen Datenverlust */
  static char      checkname[255];   /* Kopie des Dateinamens zum Vergleich
                                      * mit einer neuen Eingabe */
         char      *name;            /* Dateiname */

  switch (messageinfos->reason)
  {
         case XmCR_OK: /* Sicherung */
              if (speichernAlsText == NULL) {
                 speichernAlsText =
                      XmSelectionBoxGetChild(Widgetliste[FILE_NAME_PROMPT],
                                             XmDIALOG_TEXT);
                 warnedOfDataLoss = False;
                 strcpy(checkname, SPACE);
                 }
       
              if ((name = XmTextGetString(speichernAlsText))) {

                 if (strcmp(checkname, name) == 0) {
                    /* erneute Eingabe des gleichen Dateinamens */
                    warnedOfDataLoss = True;
                    }
                 else {
                      strcpy(checkname, name);
                      warnedOfDataLoss = False;
                      }
                   
                 if (!*name || !name) { /* keine Eingabe */
                    XBell(display, 0);
                    wprint(missingFileEntry);
                    XtManageChild(Widgetliste[FILE_NAME_PROMPT]);
                    }
                 else {
                      if (strstr(name,".")) { /* Dateiname enthaelt Endung */
                         XBell(display, 0); 
                         wprint(invalid_filename);
                         XtManageChild(Widgetliste[FILE_NAME_PROMPT]);
                         }
                      else {
                           if ((!warnedOfDataLoss) && (file_exists(name))) {
                              /* der Benutzer wurde noch nicht ueber einen
                               * moeglichen Datenverlust informiert und
                               * eine Datei gleichen Namens existiert 
                               * im Arbeitsverzeichnis  */
                              XBell(display, 0);
                              warnedOfDataLoss = True;
                              wprint(lossOfData);
                              XtManageChild(Widgetliste[FILE_NAME_PROMPT]);
                              }
                           else { /* Datei abspeichern */
                                getDrawingsize(&xsize, &ysize);
                                strcpy(data.filename_VOT, name);
                                strcpy(data.filename_DRW, name);
                                if (!saveGraphicObjects(name,
                                                       data.numb_quor, 
                                                       xsize, ysize)) {
                                   /* Fehler beim Abspeichern */
                                   XBell(display, 0);
                                   wprint(savingfailed);
                                   }
                                else { /* erfolgreiche Sicherung */
                                     clear_GraphicObjects();
                                     clearArea();
                                     redrawArea(ptrGraphicObjectList);
                                     XtUnmapWidget(Widgetliste[EDIT_BUTTONS]);

                                     if ((warnedOfDataLoss) && 
                                         (file_exists(name)))
                                        wprint(lostData, name);
                                     else wprint(saved_filename, name);

                                     XtSetSensitive(
                                          Widgetliste[NEW_TREE_BUTTON], True);
                                     XtSetSensitive(
                                          Widgetliste[OPEN_NET_BUTTON], True);
                                     XtSetSensitive(
                                          Widgetliste[OPEN_TREE_BUTTON], True);
                                     XtSetSensitive(
                                          Widgetliste[CHECK_VQM_BUTT], False);
                                     }
                                warnedOfDataLoss = False;
                                } 
                           }
                      }
                 }     
                 break;

         case XmCR_CANCEL: /* Abbruch der Sicherung */
              XtSetSensitive(Widgetliste[NEW_TREE_BUTTON], True);
              XtSetSensitive(Widgetliste[OPEN_NET_BUTTON], True);
              XtSetSensitive(Widgetliste[OPEN_TREE_BUTTON], True);
              XtSetSensitive(Widgetliste[CHECK_VQM_BUTT], True);
              break;

         default:
              XBell(display, 0);
              wprint(unknown_event, "main.c : store_freeGraphicCB");
              break; 
  }
  XtSetSensitive(Widgetliste[SAVE_YN_BUTTON], False);
}

void getDrawingsize(xsize, ysize)
     Dimension *xsize, *ysize;
/*
 * AU: Groesse des Arbeitsbereiches ermitteln
 */
{
 XtVaGetValues(Widgetliste[DRAW_AREA],
               XmNwidth, xsize,
               XmNheight, ysize, 
               NULL);
}

void checkDrawingsize()
/*
 * AU: Arbeitsbereich an die Groesse einer Darstellung angleichen
 * ST: getDrawingsize()
 */
{  
 static Dimension currentxsize, currentysize; /* gegenwaertige Groesse */

 getDrawingsize(&currentxsize, &currentysize);
  
 if ((data.draw_xsize > currentxsize) || (data.draw_ysize > currentysize)) {
    /* die Darstellung erfordert eine Anpassung des Arbeitsbereiches */
    XtVaSetValues(Widgetliste[DRAW_AREA],
                  XmNheight, Maximum(data.draw_xsize, currentxsize),
                  XmNwidth, Maximum(data.draw_ysize, currentysize), 
                  NULL);
    XBell(display, 0);
    wprint(resizeDrawArea);
    }                            
}



/*ARGSUSED*/
void exposeCB(widget, client_data, call_data)
     Widget  widget;
     caddr_t client_data, call_data;
/*
 * AU: Zeichnung des Arbeitsbereichs wieder herstellen
 * ST: redrawArea()
 *     delDependingExposeEvents()  
 */
{
  /* Arbeitsflaeche wieder zeichnen */
  redrawArea(ptrGraphicObjectList);

  /* direkt folgende Aufforderungen ignorieren */
  delDependingExposeEvents();
}

/*ARGSUSED*/
void showResultsCB(widget, client_data, call_data)
     Widget  widget;
     caddr_t client_data, call_data; 
/*
 * AU: Anzeige der Resultate beenden
 */ 
{
  XtPopdown(Widgetliste[PRES_RES_SHELL]); 
} 


Bool validNodeName(nameOfNode)
    char nameOfNode[50];
/*
 * AU: Bezeichnung eines virtuellen Rechners pruefen und Liste seiner 
 *     Repraesentanten aufbauen
 * ST: correspondingNetNode()
 *     ascending()
 *     mark_version()
 */
{
 char buffer[50]; /* Rechnerbezeichnung ohne das fuehrende 'R' */
 char *word, *ptrchr;
 char name[50], text[50];
 int  length, following_number, current_number = 0, i = 0;
 GRAPHIC_IDENT *identTemp, *identFront, *identTail; /* Zeiger auf 
                                                     * Repraesentanten */

 identTemp = NULL;
 identFront = NULL;
 identTail = NULL;

 if ((nameOfNode[0] == 'R') && (strlen(nameOfNode) <= MAX_NODENAME_LENGTH)) {
    /* Start mit 'R' und Beachtung der Eingabelaenge */   
    nameOfNode = nameOfNode + 1;
    strcpy(buffer, nameOfNode);
    length = strlen(buffer);                       

    /* Pruefung, ob die Zeichenkette aus einer durch ',' oder '-' unterbrochene
     * Folge von  Ziffern besteht - alle anderen Zeichen werden abgewiesen 
     */ 

    while (i <= length) {   
          if (isdigit(buffer[i]) != 0) { /* Ziffern ueberlesen */  
             do i++;   /* lies bis zum ersten Zeichen ungleich einer Ziffer */   
             while (isdigit(buffer[i]) != 0);   
             }   
          else {
               wprint(invalidname3); /* es wurde eine Ziffer erwartet */
               XBell(display, 0);
               return False;
               }   

          if((buffer[i] == ',') || (buffer[i] == '-') || (i >= length))
             i++; /* Ziffern sind durch ',' oder '-' unterbrochen */
          else {  
               wprint(invalidname4); /* es wurde ',' oder '-' erwartet */
               XBell(display, 0);
               return False;
               }
          }
    /*-  Pruefung, ob die Rechnernamen aufsteigend sortiert vergeben wurden:
     * Dazu wird der Name in mehrere "Token" (durch ',' und '-' separiert)
     * aufgespalten, jedes Token in eine Zahl uebersetzt und mit seinem
     * Vorgaenger auf aufsteigende Werte verglichen
     * 
     * - Pruefung, ob alle in der Bezeichnung spezifizierten Rechner auch
     * im Rechnernetz physikalisch vorhanden sind:
     * Dazu werden die durch Komma aufgezaehlten Rechner und die durch einen 
     * Bindestrich angegebene Menge von Rechnern auf ihr Vorkommen im
     * Netz geprueft
     *   
     * - Aufbau einer Liste von Repraesentanten:
     * Jedes graphische Objekt besitzt eine potentielle Liste von 
     * Repraesentanten. Bei physikalischen Rechnern ist die Liste leer, 
     * erkennbar an der Belegung der Versionsnummer mit 0. Bei virtuellen 
     * Rechnern (Versionsnummer != 0) enthaelt die Liste eine sortierte 
     * Aufzaehlung der Rechner, durch die der virtuelle Rechner 
     * repraesentiert werden kann
     */


    strcpy(text, buffer);
    word = (char *)strtok(buffer, SEPCHARS); /* Zerlegung in durch ',' oder
                                              *  '-' getrennte Token */

    if (word != NULL) {
        following_number = atoi(word);
        sprintf(name, "R%s", word);

        if (!correspondingNetNode(name)) {
            /* Existenz des physikalischen Rechners im Netz pruefen */
            wprint(invalidname7, name);
            XBell(display, 0);
            return False;
           }

        /* Aufbau des ersten Repraesentanten */
        identTemp = (GRAPHIC_IDENT *)createIdent(name);
        identFront = (GRAPHIC_IDENT *)identTemp;
        identTail = (GRAPHIC_IDENT *)identTemp;
        current_number = following_number;
       }

    word = (char *)strtok(NULL, SEPCHARS); /* Zerlegung in durch ',' oder
                                            *  '-' Token */

    while (word != NULL) {
          following_number = atoi(word); /* in eine Zahl umwandeln */
          ptrchr = (char *)strstr(text, word);

          if (ptrchr != NULL) ptrchr = ptrchr - 1;
          else                return False;

          if (*ptrchr == ',') {
             /* 
              * Aufzaehlung durch Komma getrennt 
              */

             if (!ascending(current_number, following_number)) {
                 /* Bezeichnungen sind nicht aufsteigend vergeben */
                 wprint(invalidname2);
                 XBell(display, 0);
                 return False;
                }
             else {
                  sprintf(name, "R%s", word);

                  if (!correspondingNetNode(name)) {
                      /* Existenz des physikalischen Rechners im Netz pruefen */
                      wprint(invalidname7, name);
                      XBell(display, 0);
                      return False;
                     }

                  /* Aufbau eines Repraesentanten */
                  identTemp = (GRAPHIC_IDENT *)createIdent(name);
                  identTail->nextListIdent = (GRAPHIC_IDENT *)identTemp;
                  identTail = (GRAPHIC_IDENT *)identTemp;
                  }
             }
          else {
               if (*ptrchr == '-') {
                  /* 
                   * Aufzaehlung ueberspannt einen Bereich 
                   */

                  if (!ascending(current_number, following_number)) {
                     /* Bezeichnungen sind nicht aufsteigend vergeben */
                     wprint(invalidname2);
                     XBell(display, 0);
                     return False; 
                     }

                  for (i = current_number + 1; i <= following_number; i++) {
                      /* jeder einzelne Rechner des Bereichs ist zu pruefen */
                      sprintf(name, "R%d", i);

                      if (!correspondingNetNode(name)) {
                         /* Existenz des physikalischen Rechners im Netz pruefen */
                         wprint(invalidname7, name);
                         XBell(display, 0);
                         return False;
                         }

                      /* Aufbau eines Repraesentanten */
                      identTemp = (GRAPHIC_IDENT *)createIdent(name);
                      identTail->nextListIdent = (GRAPHIC_IDENT *)identTemp;
                      identTail = (GRAPHIC_IDENT *)identTemp;
                      }
                  }
               }
          current_number = following_number;
          word = (char *)strtok(NULL, SEPCHARS); /* Ermittlung des naechsten Token */
          }
       
    data.node_ident = identFront; /* Liste der Repraesentanten global
                                   * verfuegbar */
 
    data.node_version = mark_version(identFront); /* Versionsnummer des
                                                   * virtuellen Rechners
                                                   * ermitteln */
   
    if (data.node_version == 10) {
       /* nur Versionsnummern bis 10 werden unterstuetzt */
       wprint(limitedversion);
       XBell(display, 0);
       return False;
       }
 
    if (data.node_version > 0)
       wprint(nextVersion); 

    return True; 
    }

 wprint(invalidname1);
 XBell(display, 0); 
 return False; 
}


/*ARGSUSED*/
void resultNodeNameCB(widget, clientDaten, SBinfos)
     Widget                       widget;
     caddr_t                      clientDaten;
     XmSelectionBoxCallbackStruct *SBinfos;
/*
 * AU: Verarbeitung der Benutzerreaktion auf eine ungueltige Bezeichnung
 *     fuer einen virtuellen Rechner
 * ST: setNodeListSensitive() 
 */
{
  switch (SBinfos->reason)
  {
         case XmCR_OK : /* Eingabe einer neuen Bezeichnung */        
              XtManageChild(Widgetliste[NODE_NAME_PROMPT]);
              break;
               
         case XmCR_CANCEL : /* Abbruch der Eingabe eines virtuellen Rechners */
              setNodeListSensitive(True);
              break;

         default:
              XBell(display, 0);
	      wprint(unknown_event, "main.c : resultNodeNameCB");
	      break;				   
  }
}



void extractListString(add_node, returnString)
     NODE *add_node;     /* Rechner der Netztopologie */
     char *returnString; /* zurueckzugebende Zeichenkette fuer die
                          * Darstellung als Listenelement */ 
/*
 * AU: Konstruktion der Bezeichnung eines Rechners in der Auswahlliste 
 * ST: correspondingGraphicNode()
 * NB: Rueckgabe Zeiger auf eine Zeichenkette
 */
{
  char         cstring[40];
  char         *stern = LIST_MARK;
  GRAPHIC_NODE *nodeadress = NULL;

  /* Feststellung, ob der Rechner bereits als Rechner der 
   * Voting-Struktur eingetragen wurde */
  nodeadress = (GRAPHIC_NODE *)correspondingGraphicNode(add_node->ident);

  if (nodeadress != NULL) {
     /* Rechner ist als bereits ausgewaehlt zu markieren */
     strcpy(cstring, stern);
     strcat(cstring, add_node->ident);
     }
  else strcpy(cstring, add_node->ident);

  strcat(cstring, SPACE);
  strcat(cstring, add_node->name);
  strcpy(returnString, cstring);
}



void ActualPercent(endloop, i, service)
     double endloop;
     int    i, service;
/*
 * AU: Aktuellen Bearbeitungsstand je Operation ausgeben
 */
{
  double prozent = (i * 100) / endloop;

  fprintf(stderr, statusPercent, data.host, prozent, service + 1);
}



Bool ascending(firstnumber, nextnumber)
     int firstnumber, nextnumber; /* zu vergleichende Zahlen */
/*
 * AU: Pruefung, ob eine Zahl einen  aufsteigenden Werte besitzt
 * NB: Ist die zweite Zahl groesser ist der Rueckgabewert True, sonst
 *     False
 */
{
  if (firstnumber < nextnumber ) return True;
  else                           return False;
}


 
void printUsage()
/*
 * AU: Ausgabe eines Hinweises auf den Gebrauch der Kommandozeilenoptionen
 */
{
   fprintf(stderr, usageText);
   exit(BAD);
}


void printStatus()
/*
 * AU: Ausgabe der gegenwaertigen Groesse des Arbeitsbereiches
 * ST: getDrawingsize()
 */
{
   Dimension graphikWidth, graphikHeight; /* Groesse Arbeitsbereich */

   getDrawingsize(&graphikWidth, &graphikHeight);
   fprintf(stderr, statusText, graphikWidth, graphikHeight);
}

/***************************************************************************/
/*                    Ende der Funktionen                                  */
/***************************************************************************/


